{ $Id: sortgrid.pas 159 2012-06-15 08:26:02Z Heiner $

 *****************************************************************************
 
                               sortgrid.pas
                               ------------
          An interface extending tStringGrid for Delphi or Lazarus.
               Delphi tStringGrid is not capable of sorting.                        .
             Lazarus tStringGrid is already capable of sorting,
                     but natively only by Strings.
               tSortGrid allows sorting by Numbers or DateTime.
                     Initial Revision : May 2011

 *****************************************************************************

 This file is distributed under the Library GNU General Public License
 (see the file COPYING.LGPL) with the following modification:

 As a special exception, the copyright holder of this file gives you
 permission to link this file with independent modules to produce an
 executable, regardless of the license terms of these independent modules,
 and to copy and distribute the resulting executable under terms of your choice,
 provided that you also meet, for each linked independent module, the terms
 and conditions of the license of that module. An independent module is a
 module which is not derived from or based on this file. If you modify this
 file, you may extend this exception to your version of the file, but
 you are not obligated to do so. If you do not wish to do so, delete this
 exception statement from your version.

 In order to receive a copy of the file COPYING.LGPL, contact:
      Free Software Foundation, Inc.,
      675 Mass Ave
      Cambridge, MA  02139
      USA

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 *****************************************************************************

TSortGrid for Lazarus or Delphi
Copyright (C) 2011 Heinrich Wolf.
http://www.Wolf-Fuerth.de/
}

unit sortgrid;

{$ifdef fpc}
  {$mode objfpc}{$H+}
{$endif}

interface

uses Classes,
     SysUtils,
     {$ifdef fpc}
       LResources,
     {$else}
       WinTypes,
     {$endif}
     Graphics,
     Forms,
     Controls,
     Grids,
     wolfutil;

type tMayContinue         = function : Boolean of Object;
     tCompare             = function (a, b : LongInt) : Integer of Object;
     tSwap                = procedure(a, b : LongInt) of Object;
     tFormatType          = (sgfString, sgfText, sgfInteger, sgfFloat,
                             sgfDateTime);
     {$ifndef fpc}
       tOnCompareCells    = procedure(Sender: TObject;
                                      ACol, ARow, BCol, BRow: LongInt;
                                      var Result: integer) of object;
       tSortOrder         = (soAscending, soDescending);
     {$endif}
     tOnColRowExchanged   = procedure(Sender: TObject; IsColumn: Boolean;
                                      sIndex, tIndex: LongInt) of Object;

     tGridSort            = Class(tSort)
       fMayContinue       : tMayContinue;
       fCompare           : tCompare;
       fSwap              : tSwap;
       Constructor          Create(GridMayContinue : tMayContinue;
                                   GridCompare     : tCompare;
                                   GridSwap        : tSwap);
       function             MayContinue : Boolean; override;
       function             Compare(a, b : LongInt) : Integer; override;
       procedure            Swap(a, b : LongInt); override;
     end;

     tSortArray           = Array[0..0] of tSortOrder;
     pSortArray           = ^tSortArray;
     tFormatArray         = Array[0..0] of tFormatType;
     pFormatArray         = ^tFormatArray;
     tAlignArray          = Array[0..0] of tAlignment;
     pAlignArray          = ^tAlignArray;

     tSortList          = class
     private
       fList            : pSortArray;
       fCount           : LongInt;
       function           GetSort(Pos : LongInt) : tSortOrder;
       procedure          SetSort(Pos : LongInt; NewVal : tSortOrder);
       procedure          SetCount(NewCount : LongInt);
     public
       constructor        Create;
       destructor         Destroy; override;
       property           Count : LongInt write SetCount;
       property           Items[Pos : LongInt] : tSortOrder
                            read GetSort write SetSort; default;
       procedure          Exchange(Index, WithIndex : LongInt);
     end;

     tFormatList        = class
     private
       fList            : pFormatArray;
       fCount           : LongInt;
       function           GetFormat(Pos : LongInt) : tFormatType;
       procedure          SetFormat(Pos : LongInt; NewVal : tFormatType);
       procedure          SetCount(NewCount : LongInt);
     public
       constructor        Create;
       destructor         Destroy; override;
       property           Count : LongInt write SetCount;
       property           Items[Pos : LongInt] : tFormatType
                            read GetFormat write SetFormat; default;
       procedure          Exchange(Index, WithIndex : LongInt);
     end;

     tAlignList         = class
     private
       fList            : pAlignArray;
       fCount           : LongInt;
       function           GetAlign(Pos : LongInt) : tAlignment;
       procedure          SetAlign(Pos : LongInt; NewVal : tAlignment);
       procedure          SetCount(NewCount : LongInt);
     public
       constructor        Create;
       destructor         Destroy; override;
       property           Count : LongInt write SetCount;
       property           Items[Pos : LongInt] : tAlignment
                            read GetAlign write SetAlign; default;
       procedure          Exchange(Index, WithIndex : LongInt);
     end;

     tSortGrid            = Class(tStringGrid)
     private
       fColSortOrder      : tSortList;
       fColFormatType     : tFormatList;
       fColAlignment      : tAlignList;
       fColFormatString   : tStringList;
       fIntCells          : tLongestIntMatrix;
       fFloatCells        : tExtendedMatrix;
       fSortCols          : tLongIntList;
       fSort              : tGridSort;
       fMayContinue       : Boolean;
       fShowArrow         : Boolean;
       fIsSorted          : Boolean;
       fIsSorting         : Boolean;
       fOnCompareCells    : tOnCompareCells;
       fOnColRowExchanged : tOnColRowExchanged;
       procedure            CleanUp;
       function             GetGridColCount : LongInt;
       procedure            NewGridColCount(OldCount : LongInt;
                                            var NewCount : LongInt);
       procedure            SetGridColCount(Count : LongInt);
       function             GetGridRowCount : LongInt;
       procedure            SetGridRowCount(Count : LongInt);
       function             GetMayContinue : Boolean;
       function             GetColSortOrder(GetSortCol : LongInt) : TSortOrder;
       procedure            SetColSortOrder(NewSortCol : LongInt;
                                            NewOrder   : TSortOrder);
       function             GetColFormatType(GetFormatCol : LongInt) : tFormatType;
       procedure            SetColFormatType(NewFormatCol : LongInt;
                                            FormatType : tFormatType);
       function             GetColFormatString(GetFormatCol : LongInt) : String;
       procedure            SetColFormatString(NewFormatCol : LongInt;
                                               FormatString : String);
       function             GetColAlignment(GetAlignCol : LongInt) : tAlignment;
       procedure            SetColAlignment(NewAlignCol : LongInt;
                                            Alignment : tAlignment);
       function             GetCell(ACol, ARow : LongInt) : String;
       procedure            SetCell(ACol, ARow : LongInt; Content : String);
       function             GetSortColCount : LongInt;
       function             GetSortCol(Pos : LongInt) : LongInt;
       procedure            GridColsDeleted(Index, Count : LongInt);
       procedure            GridColsExchanged(Index, WithIndex : LongInt);
       procedure            GridRowsExchanged(Index, WithIndex : LongInt);
       procedure            GridColsInserted(AtCol, Count : LongInt);
       procedure            GridColMoved(FromIndex, ToIndex : LongInt);
       procedure            SwapRows(RowA, RowB : LongInt);
       procedure            SetUnsorted;
       procedure            CheckSortCol(ACol : LongInt);
       function             CompareRows(RowA, RowB : LongInt) : Integer;
     protected
       procedure            DrawCell(ACol, ARow: Longint; ARect: TRect;
                                     AState: TGridDrawState); override;
       procedure            SetEditText(ACol, ARow: Longint;
                                        const Value: String); override;
       {$ifdef fpc}
         function           DoCompareCells(Acol, ARow, Bcol, BRow: LongInt) :
                              Integer; override;
         procedure          ColRowDeleted(IsColumn: Boolean; index: LongInt);
                              override;
         procedure          ColRowExchanged(IsColumn: Boolean;
                                            index, WithIndex: LongInt);
                              override;
         procedure          ColRowInserted(IsColumn: boolean; index: LongInt);
                              override;
         procedure          ColRowMoved(IsColumn: Boolean;
                                        FromIndex, ToIndex: LongInt); override;
       {$else}
         procedure          ColumnMoved(FromIndex, ToIndex : LongInt); override;
         procedure          RowMoved(FromIndex, ToIndex : LongInt); override;
         function           DelphiCompareCells(Acol, ARow, Bcol, BRow: LongInt) :
                              Integer;
         function           DoCompareCells(Acol, ARow, Bcol, BRow: LongInt) :
                              Integer;
         procedure          ExchangeColRow(IsColumn : Boolean;
                                           Index, WithIndex : LongInt);
       {$endif}
     public
       constructor          Create(AOwner: TComponent); override;
       destructor           Destroy; override;
       property             MayContinue : Boolean
                              read GetMayContinue write fMayContinue;
       property             ColSortOrder[SortCol : LongInt] : TSortOrder
                              read GetColSortOrder write SetColSortOrder;
       property             ColFormatType[FormatCol : LongInt] : tFormatType
                              read GetColFormatType write SetColFormatType;
       property             ColAlignment[AlignCol : LongInt] : tAlignment
                              read GetColAlignment write SetColAlignment;
       property             ColFormatString[FormatCol : LongInt] : String
                              read GetColFormatString write SetColFormatString;
       property             Cells[ACol, ARow : LongInt] : String
                              read GetCell write SetCell;
       property             IsSorted : Boolean read fIsSorted;
       property             SortColCount : LongInt read GetSortColCount;
       property             SortCols[i : LongInt] : LongInt read GetSortCol;
       function             DefaultCompareCells(
                              Acol, ARow, Bcol, BRow: LongInt) : Integer;
       procedure            SortRows(FromRow,   ToRow : LongInt);
       procedure            DeleteRows(FromRow, ToRow : LongInt);
       procedure            InsertRows(AtRow,   Count : LongInt);
       procedure            DeleteCols(FromCol, ToCol : LongInt);
       procedure            InsertCols(AtCol,   Count : LongInt);
       procedure            ExactWidthCell(ACol, ARow : LongInt);
       procedure            ExactWidthRow(ARow : LongInt);
       procedure            MinWidthCell(ACol, ARow : LongInt);
       procedure            MinWidthRow(ARow : LongInt);
       function             RequiredWidthCol(ACol : LongInt) : LongInt;
       procedure            MinWidthCol(ACol : LongInt);
       procedure            ExactWidthCol(ACol : LongInt);
       procedure            FillWidth(ACol : LongInt);
       procedure            ClearSortCols;
       procedure            AddSortCol(NewSortCol : LongInt);
     published
       property             ColCount : LongInt
                              read GetGridColCount write SetGridColCount;
       property             RowCount : LongInt
                              read GetGridRowCount write SetGridRowCount;
       property             ShowArrow : Boolean
                              read fShowArrow      write fShowArrow;
       {$ifndef fpc}
         property           OnCompareCells : tOnCompareCells
                              read fOnCompareCells write fOnCompareCells;
         property           OnColRowExchanged : tOnColRowExchanged
                              read fOnColRowExchanged write fOnColRowExchanged;
       {$endif}
     end;

procedure Register;

implementation

const
{$ifdef fpc}
  MinWidth = 8;
{$else}
  MinWidth = 4;
{$endif}

procedure Register;
  begin
    RegisterComponents('Sort Grid', [tSortGrid]);
  end;

constructor tSortList.Create;
  begin
    fList  := Nil;
    fCount := 0;
  end;

destructor tSortList.Destroy;
  begin
    if fList <> Nil then
      FreeMem(fList, fCount * sizeof(TSortOrder));
    fList  := Nil;
    fCount := 0;
  end;

function tSortList.GetSort(Pos : LongInt) : TSortOrder;
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tSortList.GetSort(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        GetSort := fList^[Pos];
      {$R+}
  end;

procedure tSortList.SetSort(Pos : LongInt; NewVal : TSortOrder);
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tSortList.SetSort(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        fList^[Pos] := NewVal;
      {$R+}
  end;

procedure tSortList.SetCount(NewCount : LongInt);
  var     NewList : pSortArray;
          i, j    : LongInt;
  begin
    if NewCount < 0 then
      Raise Exception.Create('invalid tSortList.SetCount('
                             + IntToStr(NewCount) + ')')
    else
      if NewCount <> fCount then
        begin
          GetMem(NewList, NewCount * sizeof(TSortOrder));
          if NewList = Nil then
            Raise Exception.Create('out of memory in tSortList.SetCount')
          else
            begin
              if NewCount > fCount then
                j := fCount   - 1
              else
                j := NewCount - 1;
              {$R-}
                for i := 0 to j do
                  NewList^[i] := fList^[i];
                for i := fCount to NewCount - 1 do
                  NewList^[i] := soAscending;
              {$R+}
              if fList <> Nil then
                FreeMem(fList, fCount * sizeof(TSortOrder));
              fList  := NewList;
              fCount := NewCount;
            end;
        end;
  end;

procedure tSortList.Exchange(Index, WithIndex : LongInt);
  var     Tmp : TSortOrder;
  begin
    if (Index < 0) or (Index >= fCount) then
      Raise Exception.Create('tSortList.Exchange(Index '
                             + IntToStr(Index) + ' out of range)')
    else
      if (WithIndex < 0) or (WithIndex >= fCount) then
        Raise Exception.Create('tSortList.Exchange(WithIndex '
                               + IntToStr(WithIndex) + ' out of range)')
      else
        begin
          {$R-}
            Tmp               := fList^[Index];
            fList^[Index]     := fList^[WithIndex];
            fList^[WithIndex] := Tmp;
          {$R+}
        end;
  end;

constructor tFormatList.Create;
  begin
    fList  := Nil;
    fCount := 0;
  end;

destructor tFormatList.Destroy;
  begin
    if fList <> Nil then
      FreeMem(fList, fCount * sizeof(tFormatType));
    fList  := Nil;
    fCount := 0;
  end;

function tFormatList.GetFormat(Pos : LongInt) : tFormatType;
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tFormatList.GetFormat(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        GetFormat := fList^[Pos];
      {$R+}
  end;

procedure tFormatList.SetFormat(Pos : LongInt; NewVal : tFormatType);
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tFormatList.SetFormat(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        fList^[Pos] := NewVal;
      {$R+}
  end;

procedure tFormatList.SetCount(NewCount : LongInt);
  var     NewList : pFormatArray;
          i, j    : LongInt;
  begin
    if NewCount < 0 then
      Raise Exception.Create('invalid tFormatList.SetCount('
                             + IntToStr(NewCount) + ')')
    else
      if NewCount <> fCount then
        begin
          GetMem(NewList, NewCount * sizeof(tFormatType));
          if NewList = Nil then
            Raise Exception.Create('out of memory in tFormatList.SetCount')
          else
            begin
              if NewCount > fCount then
                j := fCount   - 1
              else
                j := NewCount - 1;
              {$R-}
                for i := 0 to j do
                  NewList^[i] := fList^[i];
                for i := fCount to NewCount - 1 do
                  NewList^[i] := sgfText;
              {$R+}
              if fList <> Nil then
                FreeMem(fList, fCount * sizeof(tFormatType));
              fList  := NewList;
              fCount := NewCount;
            end;
        end;
  end;

procedure tFormatList.Exchange(Index, WithIndex : LongInt);
  var     Tmp : tFormatType;
  begin
    if (Index < 0) or (Index >= fCount) then
      Raise Exception.Create('tFormatList.Exchange(Index '
                             + IntToStr(Index) + ' out of range)')
    else
      if (WithIndex < 0) or (WithIndex >= fCount) then
        Raise Exception.Create('tFormatList.Exchange(WithIndex '
                               + IntToStr(WithIndex) + ' out of range)')
      else
        begin
          {$R-}
            Tmp               := fList^[Index];
            fList^[Index]     := fList^[WithIndex];
            fList^[WithIndex] := Tmp;
          {$R+}
        end;
  end;

constructor tAlignList.Create;
  begin
    fList  := Nil;
    fCount := 0;
  end;

destructor tAlignList.Destroy;
  begin
    if fList <> Nil then
      FreeMem(fList, fCount * sizeof(tAlignment));
    fList  := Nil;
    fCount := 0;
  end;

function tAlignList.GetAlign(Pos : LongInt) : tAlignment;
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tAlignList.GetAlign(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        GetAlign := fList^[Pos];
      {$R+}
  end;

procedure tAlignList.SetAlign(Pos : LongInt; NewVal : tAlignment);
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tAlignList.SetAlign(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        fList^[Pos] := NewVal;
      {$R+}
  end;

procedure tAlignList.SetCount(NewCount : LongInt);
  var     NewList : pAlignArray;
          i, j    : LongInt;
  begin
    if NewCount < 0 then
      Raise Exception.Create('invalid tAlignList.SetCount('
                             + IntToStr(NewCount) + ')')
    else
      if NewCount <> fCount then
        begin
          GetMem(NewList, NewCount * sizeof(tAlignment));
          if NewList = Nil then
            Raise Exception.Create('out of memory in tAlignList.SetCount')
          else
            begin
              if NewCount > fCount then
                j := fCount   - 1
              else
                j := NewCount - 1;
              {$R-}
                for i := 0 to j do
                  NewList^[i] := fList^[i];
                for i := fCount to NewCount - 1 do
                  NewList^[i] := taLeftJustify;
              {$R+}
              if fList <> Nil then
                FreeMem(fList, fCount * sizeof(tAlignment));
              fList  := NewList;
              fCount := NewCount;
            end;
        end;
  end;

procedure tAlignList.Exchange(Index, WithIndex : LongInt);
  var     Tmp : tAlignment;
  begin
    if (Index < 0) or (Index >= fCount) then
      Raise Exception.Create('tAlignList.Exchange(Index '
                             + IntToStr(Index) + ' out of range)')
    else
      if (WithIndex < 0) or (WithIndex >= fCount) then
        Raise Exception.Create('tAlignList.Exchange(WithIndex '
                               + IntToStr(WithIndex) + ' out of range)')
      else
        begin
          {$R-}
            Tmp               := fList^[Index];
            fList^[Index]     := fList^[WithIndex];
            fList^[WithIndex] := Tmp;
          {$R+}
        end;
  end;

Constructor tGridSort.Create(GridMayContinue : tMayContinue;
                             GridCompare     : tCompare;
                             GridSwap        : tSwap);
  begin
    fMayContinue := GridMayContinue;
    fCompare     := GridCompare;
    fSwap        := GridSwap;
  end;

function tGridSort.MayContinue : Boolean;
  begin
    {$ifdef fpc}
      MayContinue := fMayContinue();
    {$else}
      MayContinue := fMayContinue;
    {$endif}
  end;

function tGridSort.Compare(a, b : LongInt) : Integer;
  begin
    Compare := fCompare(a, b);
  end;

procedure tGridSort.Swap(a, b : LongInt);
  begin
    fSwap(a, b);
  end;

constructor tSortGrid.Create(AOwner: TComponent);
  var       i : LongInt;
  begin
    inherited Create(AOwner);
    fOnCompareCells    := Nil;
    fOnColRowExchanged := Nil;
    fIsSorted          := False;
    fIsSorting         := False;
    fMayContinue       := True;
    fShowArrow         := True;
    try
      fColSortOrder    := tSortList.Create;
    except
      fColSortOrder    := Nil;
    end;
    try
      fColFormatType   := tFormatList.Create;
    except
      fColFormatType   := Nil;
    end;
    try
      fColAlignment    := tAlignList.Create;
    except
      fColAlignment    := Nil;
    end;
    try
      fColFormatString := tStringList.Create;
    except
      fColFormatString := Nil;
    end;
    try
      fIntCells        := tLongestIntMatrix.Create;
    except
      fIntCells        := Nil;
    end;
    try
      fFloatCells      := tExtendedMatrix.Create;
    except
      fFloatCells      := Nil;
    end;
    try
      fSortCols        := tLongIntList.Create;
    except
      fSortCols        := Nil;
    end;
    try
      {$ifdef fpc}
        fSort          := tGridSort.Create(@GetMayContinue,
                                           @CompareRows,
                                           @SwapRows);
      {$else}
        fSort          := tGridSort.Create(GetMayContinue,
                                           CompareRows,
                                           SwapRows);
      {$endif}
    except
      fSort            := Nil;
    end;
    if   (fColSortOrder    = Nil)
      or (fColFormatType   = Nil)
      or (fColAlignment    = Nil)
      or (fColFormatString = Nil)
      or (fIntCells        = Nil)
      or (fFloatCells      = Nil)
      or (fSortCols        = Nil)
      or (fSort            = Nil)
    then
      begin
        CleanUp;
        raise Exception.Create('error in tSortGrid.Create');
      end;
    if fColFormatString <> Nil then
      begin
        fColSortOrder.Count  := inherited ColCount;
        fColFormatType.Count := inherited ColCount;
        fColAlignment.Count  := inherited ColCount;
        for i := 0 to inherited ColCount - 1 do
          fColFormatString.Add('');
      end;
  end;

destructor tSortGrid.Destroy;
  begin
    CleanUp;
    inherited Destroy;
  end;

procedure tSortGrid.CleanUp;
  begin
    fColSortOrder.Free;
    fColFormatType.Free;
    fColAlignment.Free;
    fColFormatString.Free;
    fIntCells.Free;
    fFloatCells.Free;
    fSortCols.Free;
    fSort.Free;
    fOnCompareCells    := Nil;
    fOnColRowExchanged := Nil;
    fIsSorted          := False;
    fIsSorting         := False;
    fMayContinue       := True;
    fShowArrow         := True;
    fColSortOrder      := Nil;
    fColFormatType     := Nil;
    fColAlignment      := Nil;
    fColFormatString   := Nil;
    fIntCells          := Nil;
    fFloatCells        := Nil;
    fSortCols          := Nil;
    fSort              := Nil;
  end;

procedure tSortGrid.SetUnsorted;
  begin
    if fIsSorted then
      begin
        fIsSorted := False;
        if fShowArrow then
          if fSortCols.Count > 0 then
            InvalidateCell(fSortCols[0], 0);
      end;
  end;

{$ifndef fpc}
  procedure tSortGrid.ExchangeColRow(IsColumn : Boolean;
                                     Index, WithIndex : LongInt);
    var     i : LongInt;
    begin
      if IsColumn then
        begin
          for i := 0 to inherited RowCount - 1 do
            Rows[i].Exchange(Index, WithIndex);
          GridColsExchanged(Index, WithIndex);
        end
      else
        begin
          for i := 0 to inherited ColCount - 1 do
            Cols[i].Exchange(Index, WithIndex);
          GridRowsExchanged(Index, WithIndex);
        end;
      if Assigned(fOnColRowExchanged) then
        fOnColRowExchanged(Self, IsColumn, Index, WithIndex);
    end;
{$endif}

procedure tSortGrid.SwapRows(RowA, RowB : LongInt);
  begin
    ExchangeColRow(False, RowA, RowB);
  end;

function tSortGrid.DefaultCompareCells(
           Acol, ARow, Bcol, BRow: LongInt) : Integer;
  begin
    DefaultCompareCells := AnsiCompareText(inherited Cells[ACol, ARow],
                                           inherited Cells[BCol, BRow]);
  end;

{$ifndef fpc}
  function tSortGrid.DelphiCompareCells(
             Acol, ARow, Bcol, BRow: LongInt) : Integer;
    var    i : Integer;
    begin
      if Assigned(fOnCompareCells) then
        begin
          fOnCompareCells(Self, ACol, ARow, BCol, BRow, i);
          DelphiCompareCells := i;
        end
      else
        DelphiCompareCells := AnsiCompareText(inherited Cells[ACol, ARow],
                                              inherited Cells[BCol, BRow]);
    end;
{$endif}

function tSortGrid.DoCompareCells(ACol, ARow, BCol, BRow: LongInt) : Integer;
  var    i,          Ascending  : Integer;
         IntACell,   IntBCell   : LongestInt;
         FloatACell, FloatBCell : Extended;
         StrACell,   StrBCell   : String;
  begin
    if ACol <> BCol then
      {$ifdef fpc}
        DoCompareCells := inherited DoCompareCells(ACol, ARow, BCol, BRow)
      {$else}
        DoCompareCells := DelphiCompareCells(ACol, ARow, BCol, BRow)
      {$endif}
    else
      begin
        case fColSortOrder[ACol] of
          soAscending : Ascending := 1
        else
          Ascending := -1;
        end;
        case fColFormatType[ACol] of
          sgfInteger :
            begin
              IntACell := fIntCells[ACol, ARow];
              IntBCell := fIntCells[BCol, BRow];
              if IntACell = IntBCell then
                DoCompareCells := 0
              else
                if IntACell > IntBCell then
                  DoCompareCells :=  Ascending
                else
                  DoCompareCells := -Ascending;
            end;
          sgfFloat,
          sgfDateTime :
            begin
              FloatACell := fFloatCells[ACol, ARow];
              FloatBCell := fFloatCells[BCol, BRow];
              if FloatACell = FloatBCell then
                DoCompareCells := 0
              else
                if FloatACell > FloatBCell then
                  DoCompareCells :=  Ascending
                else
                  DoCompareCells := -Ascending;
            end;
          sgfText :
            begin
              {$ifdef fpc}
                i := inherited DoCompareCells(ACol, ARow, BCol, BRow);
              {$else}
                i := DelphiCompareCells(ACol, ARow, BCol, BRow);
              {$endif}
              if i = 0 then
                DoCompareCells := 0
              else
                if i > 0 then
                  DoCompareCells :=  Ascending
                else
                  DoCompareCells := -Ascending;
            end
        else
          StrACell := inherited Cells[ACol, ARow];
          StrBCell := inherited Cells[BCol, BRow];
          if StrACell = StrBCell then
            DoCompareCells := 0
          else
            if StrACell > StrBCell then
              DoCompareCells :=  Ascending
            else
              DoCompareCells := -Ascending;
        end;
      end;
  end;

function tSortGrid.CompareRows(RowA, RowB : LongInt) : Integer;
  var    c : LongInt;
         i : Integer;
  begin
    i := 0;
    c := 0;
    while (i = 0) and (c < fSortCols.Count) do
      begin
        i := DoCompareCells(fSortCols[c], RowA, fSortCols[c], RowB);
        inc(c);
      end;
    CompareRows := i;
  end;

function tSortGrid.GetMayContinue : Boolean;
  begin
    Application.ProcessMessages;
    GetMayContinue := fMayContinue;
  end;

function tSortGrid.GetColSortOrder(GetSortCol : LongInt) : TSortOrder;
  begin
    GetColSortOrder := fColSortOrder[GetSortCol];
  end;

procedure tSortGrid.SetColSortOrder(NewSortCol : LongInt;
                                    NewOrder   : TSortOrder);
  var     c : LongInt;
  begin
    if fColSortOrder[NewSortCol] <> NewOrder then
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.SetColSortOrder while sorting')
      else
        begin
          for c := 0 to fSortCols.Count - 1 do
            if NewSortCol = fSortCols[c] then
              SetUnsorted;
          fColSortOrder[NewSortCol] := NewOrder;
        end;
  end;

function tSortGrid.GetColFormatType(GetFormatCol : LongInt) : tFormatType;
  begin
    GetColFormatType := fColFormatType[GetFormatCol];
  end;

procedure tSortGrid.SetColFormatType(NewFormatCol : LongInt;
                                     FormatType : tFormatType);
  var     c : LongInt;
  begin
    if fColFormatType[NewFormatCol] <> FormatType then
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.SetColFormatType while sorting')
      else
        begin
          for c := 0 to fSortCols.Count - 1 do
            if NewFormatCol = fSortCols[c] then
              SetUnsorted;
          fColFormatType[NewFormatCol] := FormatType;
        end;
  end;

function tSortGrid.GetColAlignment(GetAlignCol : LongInt) : tAlignment;
  begin
    GetColAlignment := fColAlignment[GetAlignCol];
  end;

procedure tSortGrid.SetColAlignment(NewAlignCol : LongInt;
                                    Alignment   : tAlignment);
  begin
    fColAlignment[NewAlignCol] := Alignment;
    {$ifdef Ver80}
      Invalidate;
    {$else}
      InvalidateCol(NewAlignCol);
    {$endif}
  end;

function tSortGrid.GetCell(ACol, ARow : LongInt) : String;
  begin
    GetCell := inherited Cells[ACol, ARow];
  end;

procedure tSortGrid.CheckSortCol(ACol : LongInt);
  var     c : LongInt;
  begin
    for c := 0 to fSortCols.Count - 1 do
      if fSortCols[c] = ACol then
        SetUnsorted;
  end;

procedure tSortGrid.SetCell(ACol, ARow : LongInt; Content : String);
  begin
    inherited Cells[ACol, ARow] := Content;
    CheckSortCol(ACol);
  end;

function tSortGrid.GetColFormatString(GetFormatCol : LongInt) : String;
  begin
    GetColFormatString := fColFormatString[GetFormatCol];
  end;

procedure tSortGrid.SetColFormatString(NewFormatCol : LongInt;
                                       FormatString : String);
  var     c : LongInt;
  begin
    if fColFormatString[NewFormatCol] <> FormatString then
      begin
        for c := 0 to fSortCols.Count - 1 do
          if fSortCols[c] = NewFormatCol then
            if fColFormatType[fSortCols[c]] = sgfDateTime then
              SetUnsorted;
        fColFormatString[NewFormatCol] := FormatString;
      end;
  end;

function tSortGrid.GetSortColCount : LongInt;
  begin
    GetSortColCount := fSortCols.Count;
  end;

function tSortGrid.GetSortCol(Pos : LongInt) : LongInt;
  begin
    GetSortCol := fSortCols[Pos];
  end;

procedure tSortGrid.AddSortCol(NewSortCol : LongInt);
  begin
    if (NewSortCol < 0) or (NewSortCol >= inherited ColCount) then
      raise Exception.Create(
        'tSortGrid.AddSortCol(' + IntToStr(NewSortCol) + ' out of range)')
    else
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.AddSortCol while sorting')
      else
        begin
          SetUnsorted;
          fSortCols.Add(NewSortCol);
        end;
  end;

procedure tSortGrid.ClearSortCols;
  begin
    SetUnsorted;
    fSortCols.Clear;
  end;

function tSortGrid.GetGridColCount : LongInt;
  begin
    GetGridColCount := inherited ColCount;
  end;

procedure tSortGrid.NewGridColCount(OldCount : LongInt; var NewCount : LongInt);
  var     i : LongInt;
  begin
    {$ifndef fpc}
      { Delphi TStringGrid does not allow ColCount 0 }
      if NewCount <= 0 then
        NewCount := 1;
    {$endif}
    fColSortOrder.Count  := NewCount;
    fColFormatType.Count := NewCount;
    fColAlignment.Count  := NewCount;
    for i := OldCount to NewCount - 1 do
      fColFormatString.Add('');
    for i := OldCount - 1 downto NewCount do
      fColFormatString.Delete(i);
    i := fSortCols.Count - 1;
    while i >= 0 do
      begin
        if fSortCols[i] >= NewCount then
          fSortCols.Delete(i);
        dec(i);
      end;
  end;

procedure tSortGrid.SetGridColCount(Count : LongInt);
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.SetGridColCount while sorting')
    else
      begin
        NewGridColCount(inherited ColCount, Count);
        inherited ColCount := Count;
      end;
  end;

function tSortGrid.GetGridRowCount : LongInt;
  begin
    GetGridRowCount := inherited RowCount;
  end;

procedure tSortGrid.SetGridRowCount(Count : LongInt);
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.SetGridRowCount while sorting')
    else
      inherited RowCount := Count;
  end;

procedure tSortGrid.DrawCell(ACol, ARow: Longint; ARect: TRect;
                             AState: TGridDrawState);
  const   HorzMargin   = 5;
          HalfArrow    = 3;
          ArrowWidth   = 2 * (HorzMargin + HalfArrow);
  var     p            : Array[0 .. 2] of tPoint;
          ArrowRect    : tRect;
          TextWidth,
          TextLeft,
          ArrowLeft,
          VertMargin   : LongInt;
          ArrowVisible : Boolean;
  begin
    if fSortCols.Count = 0 then
      ArrowVisible := False
    else
      ArrowVisible :=
        (aCol = fSortCols[0]) and (aRow = 0)
        and (ARect.Left + 2 * (HalfArrow + HorzMargin) < GridWidth)
        and fShowArrow and fIsSorted;
    if (ColAlignment[ACol] = taLeftJustify) and not ArrowVisible then
      inherited DrawCell(ACol, ARow, ARect, AState)
    else
      begin
        if gdFixed in AState then
          Canvas.Brush.Color := clBtnFace
        else
          if (AState * [gdSelected, gdFocused] = [gdSelected, gdFocused])
             and (goDrawFocusSelected in Options) or
             (AState = [gdSelected]) 
          then
            Canvas.Brush.Color := clHighlight
          else
            Canvas.Brush.Color := clWindow;
        {$ifdef fpc}
          Canvas.Pen.Color := clWindow;
        {$else}
          Canvas.Pen.Color := Canvas.Brush.Color;
        {$endif}
        Canvas.Rectangle(ARect.Left,  ARect.Top,
                         ARect.Right, ARect.Bottom);
        if gdFixed in AState then
          begin
            Canvas.Pen.Color := clBtnHighlight;
            Canvas.MoveTo(ARect.Left,      ARect.Bottom - 1);
            Canvas.LineTo(ARect.Left,      ARect.Top);
            Canvas.LineTo(ARect.Right - 1, ARect.Top);
            {$ifdef fpc}
              Canvas.Pen.Color := cl3DDKShadow;
            {$else}
              Canvas.Pen.Color := clBtnShadow;
            {$endif}
            Canvas.LineTo(ARect.Right - 1, ARect.Bottom - 1);
            Canvas.LineTo(ARect.Left,      ARect.Bottom - 1);
          end
        else
          begin
            {$ifdef fpc}
              Canvas.Pen.Color := clSilver;
              Canvas.MoveTo(ARect.Right - 1, ARect.Top);
              Canvas.LineTo(ARect.Right - 1, ARect.Bottom - 1);
              Canvas.LineTo(ARect.Left,      ARect.Bottom - 1);
            {$endif}
          end;
        TextWidth := Canvas.TextWidth(inherited Cells[aCol, aRow]);
        case ColAlignment[ACol] of
          taCenter:
            if ArrowVisible then
              begin
                TextLeft := ARect.Left +
                  (ARect.Right - ARect.Left - TextWidth - ArrowWidth) div 2;
                if TextLeft >= ARect.Left + 1 then
                  ArrowLeft := TextLeft + TextWidth
                else
                  begin
                    ArrowLeft := ARect.Right - ArrowWidth;
                    TextLeft  := ARect.Left +
                      (ARect.Right - ARect.Left - TextWidth) div 2;
                  end;
              end
            else
              begin
                TextLeft :=
                  ARect.Left + (ARect.Right - ARect.Left - TextWidth) div 2;
                ArrowLeft := TextLeft + TextWidth;
              end;
          taRightJustify:
            begin
              TextLeft  := ARect.Right - 3 - TextWidth;
              ArrowLeft := TextLeft - ArrowWidth;
              if ArrowLeft < ARect.Left + 1 then
                ArrowLeft := ARect.Left + 1;
            end;
        else
          TextLeft  := ARect.Left + 3;
          ArrowLeft := TextLeft   + TextWidth;
        end;
        {$ifdef fpc}
          Canvas.TextRect(ARect,
                          TextLeft,
                          ARect.Top + (ARect.Bottom - ARect.Top
                                       - Canvas.TextHeight(
                                           inherited Cells[aCol, aRow]))
                                      div 2,
                          inherited Cells[aCol, aRow]);
        {$else}
          Canvas.TextRect(ARect,
                          TextLeft,
                          ARect.Top + 2,
                          inherited Cells[aCol, aRow]);
        {$endif}
        if ArrowVisible then
          begin
            if ArrowLeft > ARect.Right - ArrowWidth then
              ArrowLeft := ARect.Right - ArrowWidth;
            p[0].x := ArrowLeft + HorzMargin;
            p[2].x := p[0].x    + HalfArrow;
            p[1].x := p[2].x    + HalfArrow;
            ArrowRect.Top    := ARect.Top;
            ArrowRect.Bottom := ARect.Bottom - 1;
            ArrowRect.Left   := ArrowLeft;
            ArrowRect.Right  := ArrowLeft + ArrowWidth;
            VertMargin       := (ARect.Bottom - ARect.Top) div 4;
            case fColSortOrder[fSortCols[0]] of
              soAscending :
                begin
                  p[0].y := ARect.Top    + VertMargin;
                  p[2].y := ARect.Bottom - VertMargin;
                end;
            else
              begin
                p[0].y := ARect.Bottom - VertMargin;
                p[2].y := ARect.Top    + VertMargin;
              end;
            end;
            p[1].y := p[0].y;
            Canvas.Brush.Color := clBtnFace;
            Canvas.FillRect(ArrowRect);
            Canvas.Pen.Color   := clBlack;
            Canvas.Brush.Color := clBlack;
            Canvas.Polygon(p);
          end;
      end;
  end;

procedure tSortGrid.GridColsDeleted(Index, Count : LongInt);
  var     i : LongInt;
  begin
    {$ifdef fpc}
      for i := Index to inherited ColCount - 1 do
    {$else}
      for i := Index to inherited ColCount - 1 - Count do
    {$endif}
      begin
        fColSortOrder[i]    := fColSortOrder[i + Count];
        fColFormatType[i]   := fColFormatType[i + Count];
        fColAlignment[i]    := fColAlignment[i + Count];
        fColFormatString[i] := fColFormatString[i + Count];
      end;
    i := fSortCols.Count - 1;
    while i >= 0 do
      begin
        if fSortCols[i] >= Index then
          if fSortCols[i] >= Index + Count then
            fSortCols[i] := fSortCols[i] - Count
          else
            begin
              fSortCols.Delete(i);
              SetUnsorted;
            end;
        dec(i);
      end;
  end;

procedure tSortGrid.GridColsExchanged(Index, WithIndex : LongInt);
  var     i : LongInt;
  begin
    fColSortOrder.Exchange(Index,    WithIndex);
    fColFormatType.Exchange(Index,   WithIndex);
    fColAlignment.Exchange(Index,    WithIndex);
    fColFormatString.Exchange(Index, WithIndex);
    for i := 0 to fSortCols.Count - 1 do
      if fSortCols[i] = Index then
        fSortCols[i] := WithIndex
      else
        if fSortCols[i] = WithIndex then
          fSortCols[i] := Index;
  end;

procedure tSortGrid.GridRowsExchanged(Index, WithIndex : LongInt);
  begin
    fIntCells.Exchange(Index,   WithIndex);
    fFloatCells.Exchange(Index, WithIndex);
  end;

procedure tSortGrid.GridColsInserted(AtCol, Count : LongInt);
  var     i : LongInt;
  begin
    {$ifdef fpc}
      i := inherited ColCount;
      NewGridColCount(i - Count, i);
    {$endif}
    for i := inherited ColCount - 1 downto AtCol + Count do
      begin
        fColSortOrder[i]    := fColSortOrder[i - Count];
        fColFormatType[i]   := fColFormatType[i - Count];
        fColAlignment[i]    := fColAlignment[i - Count];
        fColFormatString[i] := fColFormatString[i - Count];
        fColFormatString[i] := fColFormatString[i - Count];
      end;
    for i := AtCol to AtCol + Count - 1 do
      begin
        fColSortOrder[i]    := soAscending;
        fColFormatType[i]   := sgfText;
        fColAlignment[i]    := taLeftJustify;
        fColFormatString[i] := '';
      end;
    for i := 0 to fSortCols.Count - 1 do
      if fSortCols[i] >= AtCol then
        fSortCols[i] := fSortCols[i] + Count;
  end;

procedure tSortGrid.GridColMoved(FromIndex, ToIndex : LongInt);
  var     i                   : LongInt;
          FromColSortOrder    : TSortOrder;
          FromColFormatType   : tFormatType;
          FromColAlignment    : tAlignment;
          FromColFormatString : String;
  begin
    FromColSortOrder    := fColSortOrder[FromIndex];
    FromColFormatType   := fColFormatType[FromIndex];
    FromColAlignment    := fColAlignment[FromIndex];
    FromColFormatString := fColFormatString[FromIndex];
    if ToIndex > FromIndex then
      begin
        for i := FromIndex to ToIndex - 1 do
          begin
            fColSortOrder[i]    := fColSortOrder[i + 1];
            fColFormatType[i]   := fColFormatType[i + 1];
            fColAlignment[i]    := fColAlignment[i + 1];
            fColFormatString[i] := fColFormatString[i + 1];
          end;
        for i := 0 to fSortCols.Count - 1 do
          if fSortCols[i] = FromIndex then
            fSortCols[i] := ToIndex
          else
            if (FromIndex < fSortCols[i])
               and (fSortCols[i] <= ToIndex)
            then
              fSortCols[i] := fSortCols[i] - 1;
      end
    else
      begin
        for i := FromIndex downto ToIndex + 1  do
          begin
            fColSortOrder[i]    := fColSortOrder[i - 1];
            fColFormatType[i]   := fColFormatType[i - 1];
            fColAlignment[i]    := fColAlignment[i - 1];
            fColFormatString[i] := fColFormatString[i - 1];
          end;
        for i := 0 to fSortCols.Count - 1 do
          if fSortCols[i] = FromIndex then
            fSortCols[i] := ToIndex
          else
            if (ToIndex <= fSortCols[i]) and (fSortCols[i] < FromIndex) then
              fSortCols[i] := fSortCols[i] + 1;
      end;
    fColSortOrder[ToIndex]    := FromColSortOrder;
    fColFormatType[ToIndex]   := FromColFormatType;
    fColAlignment[ToIndex]    := FromColAlignment;
    fColFormatString[ToIndex] := FromColFormatString;
  end;

{$ifdef fpc}
  procedure tSortGrid.ColRowDeleted(IsColumn: Boolean; index: LongInt);
    begin
      inherited ColRowDeleted(IsColumn, index);
      if IsColumn then
        GridColsDeleted(index, 1);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.ColRowDeleted while sorting')
    end;

  procedure tSortGrid.ColRowExchanged(IsColumn: Boolean;
                                      index, WithIndex: LongInt);
    begin
      inherited ColRowExchanged(IsColumn, index, WithIndex);
      if IsColumn then
        begin
          GridColsExchanged(index, WithIndex);
          if fIsSorting then
            raise Exception.Create(
              'must not tSortGrid.ColRowExchanged while sorting');
        end
      else
        GridRowsExchanged(index, WithIndex);
    end;

  procedure tSortGrid.ColRowInserted(IsColumn: boolean; index: LongInt);
    begin
      inherited ColRowInserted(IsColumn, index);
      if IsColumn then
        GridColsInserted(Index, 1);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.ColRowInserted while sorting');
    end;

  procedure tSortGrid.ColRowMoved(IsColumn: Boolean;
                                  FromIndex, ToIndex: LongInt);
    begin
      inherited ColRowMoved(IsColumn, FromIndex, ToIndex);
      if IsColumn then
        GridColMoved(FromIndex, ToIndex);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.GridColumnMoved while sorting');
    end;
{$else}
  procedure tSortGrid.ColumnMoved(FromIndex, ToIndex : LongInt);
    begin
      inherited ColumnMoved(FromIndex, ToIndex);
      GridColMoved(FromIndex, ToIndex);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.ColumnMoved while sorting');
    end;

  procedure tSortGrid.RowMoved(FromIndex, ToIndex : LongInt);
    begin
      inherited RowMoved(FromIndex, ToIndex);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.RowMoved while sorting');
    end;
{$endif}

procedure tSortGrid.SetEditText(ACol, ARow: Longint; const Value: String);
  begin
    inherited SetEditText(ACol, ARow, Value);
    CheckSortCol(ACol);
  end;

procedure tSortGrid.SortRows(FromRow, ToRow : LongInt);
  var     i, j   : LongInt;
          Format,
          Error  : String;
  begin
    if fIsSorting then
      raise Exception.Create('tSortGrid.SortRows: already sorting')
    else
      if fSortCols.Count > 0 then
        begin
          fMayContinue := True;
          fIsSorting   := True;
          Error        := '';
          if FromRow > ToRow then
            begin
              i       := FromRow;
              FromRow := ToRow;
              ToRow   := i;
            end;
          try
            for i := 0 to Inherited ColCount - 1 do
              case fColFormatType[i] of
                sgfInteger:
                  begin
                    fIntCells.AddInts(ToRow + 1);
                    fFloatCells.AddNil;
                    for j := FromRow to ToRow do
                      if Trim(inherited Cells[i, j]) = '' then
                        fIntCells[i, j] := 0
                      else
                        try
                          fIntCells[i, j] :=
                            DezToInt(inherited Cells[i, j]);
                        except
                          fIntCells[i, j] := 0;
                        end;
                  end;
                sgfFloat:
                  begin
                    fFloatCells.AddExtended(ToRow + 1);
                    fIntCells.AddNil;
                    for j := FromRow to ToRow do
                      if Trim(inherited Cells[i, j]) = '' then
                        fFloatCells[i, j] := 0
                      else
                        try
                          fFloatCells[i, j] :=
                            StrToFloat(StringReplace(inherited Cells[i, j],
                                                     GetThousandSeparator,
                                                     '', [rfReplaceAll]));
                        except
                          fFloatCells[i, j] := 0;
                        end;
                  end;
                sgfDateTime:
                  begin
                    fFloatCells.AddExtended(ToRow + 1);
                    fIntCells.AddNil;
                    Format := Trim(ColFormatString[i]);
                    if Format = '' then
                      for j := FromRow to ToRow do
                        if Trim(inherited Cells[i, j]) = '' then
                          fFloatCells[i, j] := 0
                        else
                          try
                            fFloatCells[i, j] :=
                              StrToDateTime(inherited Cells[i, j])
                          except
                            fFloatCells[i, j] := 0;
                          end
                    else
                      for j := FromRow to ToRow do
                        if Trim(inherited Cells[i, j]) = '' then
                          fFloatCells[i, j] := 0
                        else
                          try
                            fFloatCells[i, j] :=
                              StrWithFormatToDateTime(
                                inherited Cells[i, j], Format);
                          except
                            fFloatCells[i, j] := 0;
                          end;
                  end;
              else
                begin
                  fIntCells.AddNil;
                  fFloatCells.AddNil;
                end;
              end;
            fSort.Sort(FromRow, ToRow);
          except
            on e : Exception do
              Error := e.Message;
          end;
          fIntCells.Clear;
          fFloatCells.Clear;
          fIsSorting := False;
          if fMayContinue then
            fIsSorted := True;
          InvalidateCell(fSortCols[0], 0);
          if Error > '' then
            raise Exception.Create(Error);
        end;
  end;

procedure tSortGrid.DeleteRows(FromRow, ToRow : LongInt);
  var     i, Count : LongInt;
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.DeleteRows while sorting')
    else
      begin
        if ToRow < FromRow then
          begin
            i       := ToRow;
            ToRow   := FromRow;
            FromRow := i;
          end;
        Count := ToRow - FromRow + 1;
        {$ifdef fpc}
          for i := 1 to Count do
            DeleteRow(FromRow);
        {$else}
          for i := FromRow to inherited RowCount - 1 - Count do
            Rows[i] := Rows[i + Count];
          inherited RowCount := inherited RowCount - Count;
        {$endif}
      end;
  end;

procedure tSortGrid.InsertRows(AtRow, Count : LongInt);
  var     i        : LongInt;
          {$ifdef fpc}
            r      : LongInt;
          {$else}
            MaxRow : LongInt;
          {$endif}
  begin
    if fIsSorting then
      raise Exception.Create('must not tSortGrid.InsertRows while sorting')
    else
      begin
        {$ifdef fpc}
          r := Row;
          for i := 1 to count do
            InsertColRow(False, AtRow);
          if r = AtRow then
            Row := r;
        {$else}
          MaxRow             := inherited RowCount - 1;
          inherited RowCount := inherited RowCount + Count;
          for i := MaxRow downto AtRow do
            Rows[i + Count] := Rows[i];
          for i := AtRow to AtRow + Count - 1 do
            Rows[i].Clear;
        {$endif}
      end;
  end;

procedure tSortGrid.InsertCols(AtCol, Count : LongInt);
  var     i : LongInt;
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.InsertCols while sorting')
    else
      begin
        {$ifdef fpc}
          for i := 1 to Count do
            InsertColRow(True, AtCol);
        {$else}
          SetGridColCount(inherited ColCount + Count);
          for i := inherited ColCount - 1 downto AtCol + Count do
            begin
              Cols[i]      := Cols[i - Count];
              ColWidths[i] := ColWidths[i - Count];
            end;
          for i := AtCol to AtCol + Count - 1 do
            Cols[i].Clear;
          GridColsInserted(AtCol, Count);
        {$endif}
      end;
  end;

procedure tSortGrid.DeleteCols(FromCol, ToCol : LongInt);
  var     i, Count : LongInt;
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.DeleteCols while sorting')
    else
      begin
        if ToCol < FromCol then
          begin
            i       := ToCol;
            ToCol   := FromCol;
            FromCol := i;                
          end;
        Count := ToCol - FromCol + 1;
        {$ifdef fpc}
          for i := 1 to Count do
            DeleteCol(FromCol);
        {$else}
          for i := FromCol to inherited ColCount - 1 - Count do
            begin
              Cols[i]      := Cols[i + Count];
              ColWidths[i] := ColWidths[i + Count];
            end;
          GridColsDeleted(FromCol, Count);
          SetGridColCount(inherited ColCount - Count);
        {$endif}
      end;
  end;

procedure tSortGrid.ExactWidthCell(ACol, ARow : LongInt);
  begin
    ColWidths[ACol] := Canvas.TextWidth(inherited Cells[ACol, ARow])
                       + GridLineWidth + MinWidth;
  end;

procedure tSortGrid.ExactWidthRow(ARow : LongInt);
  var i : LongInt;
  begin
    for i := 0 to inherited ColCount - 1 do
      ExactWidthCell(i, ARow);
  end;

procedure tSortGrid.MinWidthCell(ACol, ARow : LongInt);
  var w : LongInt;
  begin
    w := Canvas.TextWidth(inherited Cells[ACol, ARow])
         + GridLineWidth + MinWidth;
    if ColWidths[ACol] < w then
      ColWidths[ACol] := w;
  end;

procedure tSortGrid.MinWidthRow(ARow : LongInt);
  var i : LongInt;
  begin
    for i := 0 to inherited ColCount - 1 do
      MinWidthCell(i, ARow);
  end;

function tSortGrid.RequiredWidthCol(ACol : LongInt) : LongInt;
  var    i, w, c : LongInt;
  begin
    c := 0;
    for i := 0 to inherited RowCount - 1 do
      begin
        w := Canvas.TextWidth(inherited Cells[ACol, i])
             + GridLineWidth + MinWidth;
        if w > c then
          c := w;
      end;
    RequiredWidthCol := c;
  end;

procedure tSortGrid.MinWidthCol(ACol : LongInt);
  var w : LongInt;
  begin
    w := RequiredWidthCol(ACol);
    if ColWidths[ACol] < w then
      ColWidths[ACol] := w;
  end;

procedure tSortGrid.ExactWidthCol(ACol : LongInt);
  begin
    ColWidths[ACol] := RequiredWidthCol(ACol);
  end;

procedure tSortGrid.FillWidth(ACol : LongInt);
  var     w, i : Integer;
  begin
    w := ClientWidth;
    for i := 0 to ColCount - 1 do
      if i <> ACol then
        dec(w, ColWidths[i] + GridLineWidth);
    if ColWidths[ACol] < w then
      ColWidths[ACol] := w;
  end;

end.
